#!/usr/bin/env lua


---------------------------
-- module initialization --
---------------------------

local _G              = _G
local _VERSION        = _VERSION
local assert          = assert
local collectgarbage  = collectgarbage
local dofile          = dofile
local error           = error
local getfenv         = getfenv
local getmetatable    = getmetatable
local ipairs          = ipairs
local load            = load
local loadfile        = loadfile
local loadstring      = loadstring
local next            = next
local pairs           = pairs
local pcall           = pcall
local print           = print
local rawequal        = rawequal
local rawget          = rawget
local rawset          = rawset
local select          = select
local setfenv         = setfenv
local setmetatable    = setmetatable
local tonumber        = tonumber
local tostring        = tostring
local type            = type
local unpack          = unpack
local xpcall          = xpcall

local coroutine       = coroutine
local io              = io
local math            = math
local os              = os
local string          = string
local table           = table

local add             = table.insert

_G[...] = require("mondelefant_native")
module(...)



---------------
-- selectors --
---------------

selector_metatable = {}
selector_prototype = {}
selector_metatable.__index = selector_prototype

local function init_selector(self, db_conn)
  self._db_conn = db_conn
  self._mode = "list"
  self._fields = { sep = ", " }
  self._distinct = false
  self._distinct_on = {sep = ", ", expression}
  self._from = { sep = " " }
  self._where = { sep = " AND " }
  self._group_by = { sep = ", " }
  self._having = { sep = " AND " }
  self._combine = { sep = " " }
  self._order_by = { sep = ", " }
  self._limit = nil
  self._offset = nil
  --[[
  self._lock = nil
  self._lock_tables = { sep = ", " }
  --]]
  self._class = nil
  self._attach = nil
  return self
end

function connection_prototype:new_selector()
  return init_selector(setmetatable({}, selector_metatable), self)
end

function selector_prototype:get_db_conn()
  return self._db_conn
end

-- TODO: selector clone?

function selector_prototype:single_object_mode()
  self._mode = "object"
  return self
end

function selector_prototype:optional_object_mode()
  self._mode = "opt_object"
  return self
end

function selector_prototype:empty_list_mode()
  self._mode = "empty_list"
  return self
end

function selector_prototype:add_distinct_on(expression)
  if self._distinct then
    error("Can not combine DISTINCT with DISTINCT ON.")
  end
  add(self._distinct_on, expression)
  return self
end

function selector_prototype:set_distinct()
  if #self._distinct_on > 0 then
    error("Can not combine DISTINCT with DISTINCT ON.")
  end
  self._distinct = true
  return self
end

function selector_prototype:add_from(expression, alias, condition)
  local first = (#self._from == 0)
  if not first then
    if condition then
      add(self._from, "INNER JOIN")
    else
      add(self._from, "CROSS JOIN")
    end
  end
  if getmetatable(expression) == selector_metatable then
    if alias then
      add(self._from, {'($) AS "$"', {expression}, {alias}})
    else
      add(self._from, {'($) AS "subquery"', {expression}})
    end
  else
    if alias then
      add(self._from, {'$ AS "$"', {expression}, {alias}})
    else
      add(self._from, expression)
    end
  end
  if condition then
    if first then
      self:condition(condition)
    else
      add(self._from, "ON")
      add(self._from, condition)
    end
  end
  return self
end

function selector_prototype:add_where(expression)
  add(self._where, expression)
  return self
end

function selector_prototype:add_group_by(expression)
  add(self._group_by, expression)
  return self
end

function selector_prototype:add_having(expression)
  add(self._having, expression)
  return self
end

function selector_prototype:add_combine(expression)
  add(self._combine, expression)
  return self
end

function selector_prototype:add_order_by(expression)
  add(self._order_by, expression)
  return self
end

function selector_prototype:limit(count)
  if type(count) ~= "number" or count % 1 ~= 0 then
    error("LIMIT must be an integer.")
  end
  self._limit = count
  return self
end

function selector_prototype:offset(count)
  if type(count) ~= "number" or count % 1 ~= 0 then
    error("OFFSET must be an integer.")
  end
  self._offset = count
  return self
end

function selector_prototype:reset_fields()
  for idx in ipairs(self._fields) do
    self._fields[idx] = nil
  end
  return self
end

function selector_prototype:add_field(expression, alias, options)
  if alias then
    add(self._fields, {'$ AS "$"', {expression}, {alias}})
  else
    add(self._fields, expression)
  end
  if options then
    for i, option in ipairs(options) do
      if option == "distinct" then
        if alias then
          self:add_distinct_on('"' .. alias .. '"')
        else
          self:add_distinct_on(expression)
        end
      elseif option == "grouped" then
        if alias then
          self:add_group_by('"' .. alias .. '"')
        else
          self:add_group_by(expression)
        end
      else
        error("Unknown option '" .. option .. "' to add_field method.")
      end
    end
  end
  return self
end

function selector_prototype:join(...)  -- NOTE: alias for add_from
  return self:add_from(...)
end

function selector_prototype:from(expression, alias, condition)
  if #self._from > 0 then
    error("From-clause already existing (hint: try join).")
  end
  return self:join(expression, alias, condition)
end

function selector_prototype:left_join(expression, alias, condition)
  local first = (#self._from == 0)
  if not first then
    add(self._from, "LEFT OUTER JOIN")
  end
  if alias then
    add(self._from, {'$ AS "$"', {expression}, {alias}})
  else
    add(self._from, expression)
  end
  if condition then
    if first then
      self:condition(condition)
    else
      add(self._from, "ON")
      add(self._from, condition)
    end
  end
  return self
end

function selector_prototype:union(expression)
  self:add_combine{"UNION $", {expression}}
  return self
end

function selector_prototype:union_all(expression)
  self:add_combine{"UNION ALL $", {expression}}
  return self
end

function selector_prototype:intersect(expression)
  self:add_combine{"INTERSECT $", {expression}}
  return self
end

function selector_prototype:intersect_all(expression)
  self:add_combine{"INTERSECT ALL $", {expression}}
  return self
end

function selector_prototype:except(expression)
  self:add_combine{"EXCEPT $", {expression}}
  return self
end

function selector_prototype:except_all(expression)
  self:add_combine{"EXCEPT ALL $", {expression}}
  return self
end

function selector_prototype:set_class(class)
  self._class = class
  return self
end

function selector_prototype:attach(mode, data2, field1, field2, ref1, ref2)
  self._attach = {
    mode = mode,
    data2 = data2,
    field1 = field1,
    field2 = field2,
    ref1 = ref1,
    ref2 = ref2
  }
  return self
end

-- TODO: many-to-many relations

function selector_metatable:__tostring()
  local parts = {sep = " "}
  add(parts, "SELECT")
  if self._distinct then
    add(parts, "DISTINCT")
  elseif #self._distinct_on > 0 then
    add(parts, {"DISTINCT ON ($)", self._distinct_on})
  end
  add(parts, {"$", self._fields})
  if #self._from > 0 then
    add(parts, {"FROM $", self._from})
  end
  if #self._mode == "empty_list" then
    add(parts, "WHERE FALSE")
  elseif #self._where > 0 then
    add(parts, {"WHERE $", self._where})
  end
  if #self._group_by > 0 then
    add(parts, {"GROUP BY $", self._group_by})
  end
  if #self._having > 0 then
    add(parts, {"HAVING $", self._having})
  end
  for i, v in ipairs(self._combine) do
    add(parts, v)
  end
  if #self._order_by > 0 then
    add(parts, {"ORDER BY $", self._order_by})
  end
  if self._mode == "empty_list" then
    add(parts, "LIMIT 0")
  elseif self._mode ~= "list" then
    add(parts, "LIMIT 1")
  elseif self._limit then
    add(parts, "LIMIT " .. self._limit)
  end
  if self._offset then
    add(parts, "OFFSET " .. self._offset)
  end
  return self._db_conn:assemble_command{"$", parts}
end

function selector_prototype:try_exec()
  if self._mode == "empty_list" then
    if self._class then
      return nil, self._class:create_list()
    else
       return nil, self._db_conn:create_list()
    end
  end
  local db_error, db_result = self._db_conn:try_query(self, self._mode)
  if db_error then
    return db_error
  elseif db_result then
    if self._class then set_class(db_result, self._class) end
    if self._attach then
      attach(
        self._attach.mode,
        db_result,
        self._attach.data2,
        self._attach.field1,
        self._attach.field2,
        self._attach.ref1,
        self._attach.ref2
      )
    end
    return nil, db_result
  else
    return nil
  end
end

function selector_prototype:exec()
  local db_error, result = self:try_exec()
  if db_error then
    db_error:escalate()
  else
    return result
  end
end



-----------------
-- attachments --
-----------------

local function attach_key(row, fields)
  local t = type(fields)
  if t == "string" then
    return tostring(row[fields])
  elseif t == "table" then
    local r = {}
    for idx, field in ipairs(fields) do
      r[idx] = string.format("%q", row[field])
    end
    return table.concat(r)
  else
    error("Field information for 'mondelefant.attach' is neither a string nor a table.")
  end
end

function attach(mode, data1, data2, key1, key2, ref1, ref2)
  local many1, many2
  if mode == "11" then
    many1 = false
    many2 = false
  elseif mode == "1m" then
    many1 = false
    many2 = true
  elseif mode == "m1" then
    many1 = true
    many2 = false
  elseif mode == "mm" then
    many1 = true
    many2 = true
  else
    error("Unknown mode specified for 'mondelefant.attach'.")
  end
  local list1, list2
  if data1._type == "object" then
    list1 = { data1 }
  elseif data1._type == "list" then
    list1 = data1
  else
    error("First result data given to 'mondelefant.attach' is invalid.")
  end
  if data2._type == "object" then
    list2 = { data2 }
  elseif data2._type == "list" then
    list2 = data2
  else
    error("Second result data given to 'mondelefant.attach' is invalid.")
  end
  local hash1 = {}
  local hash2 = {}
  if ref2 then
    for i, row in ipairs(list1) do
      local key = attach_key(row, key1)
      local list = hash1[key]
      if not list then list = {}; hash1[key] = list end
      list[#list + 1] = row
    end
  end
  if ref1 then
    for i, row in ipairs(list2) do
      local key = attach_key(row, key2)
      local list = hash2[key]
      if not list then list = {}; hash2[key] = list end
      list[#list + 1] = row
    end
    for i, row in ipairs(list1) do
      local key = attach_key(row, key1)
      local matching_rows = hash2[key]
      if many2 then
        local list = data2._connection:create_list(matching_rows)
        list._class = data2._class
        row._ref[ref1] = list
      elseif matching_rows and #matching_rows == 1 then
        row._ref[ref1] = matching_rows[1]
      else
        row._ref[ref1] = false
      end
    end
  end
  if ref2 then
    for i, row in ipairs(list2) do
      local key = attach_key(row, key2)
      local matching_rows = hash1[key]
      if many1 then
        local list = data1._connection:create_list(matching_rows)
        list._class = data1._class
        row._ref[ref2] = list
      elseif matching_rows and #matching_rows == 1 then
        row._ref[ref2] = matching_rows[1]
      else
        row._ref[ref2] = false
      end
    end
  end
end



------------------
-- model system --
------------------

class_prototype.primary_key = "id"

function class_prototype:get_db_conn()
  error(
    "Method mondelefant class(_prototype):get_db_conn() " ..
    "has to be implemented."
  )
end

function class_prototype:get_qualified_table()
  if not self.table then error "Table unknown." end
  if self.schema then
    return '"' .. self.schema .. '"."' .. self.table .. '"'
  else
    return '"' .. self.table .. '"'
  end
end

function class_prototype:get_qualified_table_literal()
  if not self.table then error "Table unknown." end
  if self.schema then
    return self.schema .. '.' .. self.table
  else
    return self.table
  end
end

function class_prototype:get_primary_key_list()
  local primary_key = self.primary_key
  if type(primary_key) == "string" then
    return {primary_key}
  else
    return primary_key
  end
end

function class_prototype:get_columns()
  if self._columns then
    return self._columns
  end
  local selector = self:get_db_conn():new_selector()
  selector:set_class(self)
  selector:from(self:get_qualified_table())
  selector:add_field("*")
  selector:add_where("FALSE")
  local db_result = selector:exec()
  local connection = db_result._connection
  local columns = {}
  for idx, info in ipairs(db_result._column_info) do
    local key   = info.field_name
    local value = {
      name = key,
      type = connection.type_mappings[info.type]
    }
    columns[key] = value
    table.insert(columns, value)
  end
  self._columns = columns
  return columns
end

function class_prototype:new_selector(db_conn)
  local selector = (db_conn or self:get_db_conn()):new_selector()
  selector:set_class(self)
  selector:from(self:get_qualified_table())
  selector:add_field(self:get_qualified_table() .. ".*")
  return selector
end

function class_prototype:create_list()
  local list = self:get_db_conn():create_list()
  list._class = self
  return list
end

function class_prototype:new()
  local object = self:get_db_conn():create_object()
  object._class = self
  object._new = true
  return object
end

function class_prototype.object:try_save()
  if not self._class then
    error("Cannot save object: No class information available.")
  end
  local primary_key = self._class:get_primary_key_list()
  local primary_key_sql = { sep = ", " }
  for idx, value in ipairs(primary_key) do
    primary_key_sql[idx] = '"' .. value .. '"'
  end
  if self._new then
    local fields = {sep = ", "}
    local values = {sep = ", "}
    for key, dummy in pairs(self._dirty or {}) do
      add(fields, {'"$"', {key}})
      add(values, {'?', self[key]})
    end
    if compat_returning then  -- compatibility for PostgreSQL 8.1
      local db_error, db_result1, db_result2 = self._connection:try_query(
        {
          'INSERT INTO $ ($) VALUES ($)',
          {self._class:get_qualified_table()},
          fields,
          values,
          primary_key_sql
        },
        "list",
        {
          'SELECT currval(?)',
          self._class.table .. '_id_seq'
        },
        "object"
      )
      if db_error then
        return db_error
      end
      self.id = db_result2.id
    else
      local db_error, db_result = self._connection:try_query(
        {
          'INSERT INTO $ ($) VALUES ($) RETURNING ($)',
          {self._class:get_qualified_table()},
          fields,
          values,
          primary_key_sql
        },
        "object"
      )
      if db_error then
        return db_error
      end
      for idx, value in ipairs(primary_key) do
        self[value] = db_result[value]
      end
    end
    self._new = false
  else
    local command_sets = {sep = ", "}
    for key, dummy in pairs(self._dirty or {}) do
      add(command_sets, {'"$" = ?', {key}, self[key]})
    end
    if #command_sets >= 1 then
      local primary_key_compare = {sep = " AND "}
      for idx, value in ipairs(primary_key) do
        primary_key_compare[idx] = {
          "$ = ?",
          {'"' .. value .. '"'},
          self[value]
        }
      end
      local db_error = self._connection:try_query{
        'UPDATE $ SET $ WHERE $',
        {self._class:get_qualified_table()},
        command_sets,
        primary_key_compare
      }
      if db_error then
        return db_error
      end
    end
  end
  return nil
end

function class_prototype.object:save()
  local db_error = self:try_save()
  if db_error then
    db_error:escalate()
  end
  return self
end

function class_prototype.object:try_destroy()
  if not self._class then
    error("Cannot destroy object: No class information available.")
  end
  local primary_key = self._class:get_primary_key_list()
  local primary_key_compare = {sep = " AND "}
  for idx, value in ipairs(primary_key) do
    primary_key_compare[idx] = {
      "$ = ?",
      {'"' .. value .. '"'},
      self[value]
    }
  end
  return self._connection:try_query{
    'DELETE FROM $ WHERE $',
    {self._class:get_qualified_table()},
    primary_key_compare
  }
end

function class_prototype.object:destroy()
  local db_error = self:try_destroy()
  if db_error then
    db_error:escalate()
  end
  return self
end

function class_prototype.list:get_reference_selector(
  ref_name, options, ref_alias, back_ref_alias
)
  local ref_info = self._class.references[ref_name]
  if not ref_info then
    error('Reference with name "' .. ref_name .. '" not found.')
  end
  local selector = ref_info.selector_generator(self, options or {})
  local mode = ref_info.mode
  if mode == "mm" or mode == "1m" then
    mode = "m1"
  elseif mode == "m1" then
    mode = "1m"
  end
  local ref_alias = ref_alias
  if ref_alias == false then
    ref_alias = nil
  elseif ref_alias == nil then
    ref_alias = ref_name
  end
  local back_ref_alias
  if back_ref_alias == false then
    back_ref_alias = nil
  elseif back_ref_alias == nil then
    back_ref_alias = ref_info.back_ref
  end
  selector:attach(
    mode,
    self,
    ref_info.that_key,                   ref_info.this_key,
    back_ref_alias or ref_info.back_ref, ref_alias or ref_name
  )
  return selector
end

function class_prototype.list.load(...)
  return class_prototype.list.get_reference_selector(...):exec()
end

function class_prototype.object:get_reference_selector(...)
  local list = self._class:create_list()
  list[1] = self
  return list:get_reference_selector(...)
end

function class_prototype.object.load(...)
  return class_prototype.object.get_reference_selector(...):exec()
end


function class_prototype:add_reference(args)
  local selector_generator    = args.selector_generator
  local mode                  = args.mode
  local to                    = args.to
  local this_key              = args.this_key
  local that_key              = args.that_key
  local connected_by_table    = args.connected_by_table  -- TODO: split to table and schema
  local connected_by_this_key = args.connected_by_this_key
  local connected_by_that_key = args.connected_by_that_key
  local ref                   = args.ref
  local back_ref              = args.back_ref
  local default_order         = args.default_order
  local model
  local function get_model()
    if not model then
      if type(to) == "string" then
        model = _G
        for path_element in string.gmatch(to, "[^.]+") do
          model = model[path_element]
        end
      elseif type(to) == "function" then
        model = to()
      else
        model = to
      end
    end
    if not model or model == _G then
      error("Could not get model for reference.")
    end
    return model
  end
  self.references[ref] = {
    mode     = mode,
    this_key = this_key,
    that_key = connected_by_table and "mm_ref_" or that_key,
    ref      = ref,
    back_ref = back_ref,
    selector_generator = selector_generator or function(list, options)
      -- TODO: support tuple keys
      local options = options or {}
      local model = get_model()
      -- TODO: too many records cause PostgreSQL command stack overflow
      local ids = { sep = ", " }
      for i, object in ipairs(list) do
        local id = object[this_key]
        if id ~= nil then
          ids[#ids+1] = {"?", id}
        end
      end
      if #ids == 0 then
        return model:new_selector():empty_list_mode()
      end
      local selector = model:new_selector()
      if connected_by_table then
        selector:join(
          connected_by_table,
          nil,
          {
            '$."$" = $."$"',
            {connected_by_table},
            {connected_by_that_key},
            {model:get_qualified_table()},
            {that_key}
          }
        )
        selector:add_field(
          {
            '$."$"',
            {connected_by_table},
            {connected_by_this_key}
          },
          'mm_ref_'
        )
        selector:add_where{
          '$."$" IN ($)',
          {connected_by_table},
          {connected_by_this_key},
          ids
        }
      else
        selector:add_where{'"$" IN ($)', {that_key}, ids}
      end
      if options.order == nil and default_order then
        selector:add_order_by(default_order)
      elseif options.order then
        selector:add_order_by(options.order)
      end
      return selector
    end
  }
  if mode == "m1" or mode == "11" then
    self.foreign_keys[this_key] = ref
  end
  return self
end
