#include <lua.h>
#include <lauxlib.h>
#include <libpq-fe.h>
#include <postgres.h>
#include <catalog/pg_type.h>
#include <stdint.h>

#define MONDELEFANT_REGKEY "e449ba8d9a53d353_mondelefant"

#define MONDELEFANT_MODULE_REGKEY (MONDELEFANT_REGKEY "_module")
#define MONDELEFANT_CONN_MT_REGKEY (MONDELEFANT_REGKEY "_connection")
#define MONDELEFANT_CONN_DATA_REGKEY (MONDELEFANT_REGKEY "_connection_data")
#define MONDELEFANT_RESULT_MT_REGKEY (MONDELEFANT_REGKEY "_result")
#define MONDELEFANT_ERROROBJECT_MT_REGKEY (MONDELEFANT_REGKEY "_errorobject")
#define MONDELEFANT_CLASS_MT_REGKEY (MONDELEFANT_REGKEY "_class")
#define MONDELEFANT_CLASS_PROTO_REGKEY (MONDELEFANT_REGKEY "_class_proto")

#define MONDELEFANT_SERVER_ENCODING_ASCII 0
#define MONDELEFANT_SERVER_ENCODING_UTF8  1

typedef struct {
  PGconn *pgconn;
  int server_encoding;
} mondelefant_conn_t;

static size_t utf8_position_to_byte(const char *str, size_t utf8pos) {
  size_t bytepos;
  for (bytepos = 0; utf8pos > 0; bytepos++) {
    uint8_t c;
    c = ((const uint8_t *)str)[bytepos];
    if (!c) break;
    if (c <= 0x7f || c >= 0xc0) utf8pos--;
  }
  return bytepos;
}

#define MONDELEFANT_POSTGRESQL_BINARY_OID ((Oid)17)

static const char *mondelefant_oid_to_typestr(Oid oid) {
  switch (oid) {
    case 16: return "bool";
    case 17: return "bytea";
    case 18: return "char";
    case 19: return "name";
    case 20: return "int8";
    case 21: return "int2";
    case 23: return "int4";
    case 25: return "text";
    case 26: return "oid";
    case 27: return "tid";
    case 28: return "xid";
    case 29: return "cid";
    case 600: return "point";
    case 601: return "lseg";
    case 602: return "path";
    case 603: return "box";
    case 604: return "polygon";
    case 628: return "line";
    case 700: return "float4";
    case 701: return "float8";
    case 705: return "unknown";
    case 718: return "circle";
    case 790: return "money";
    case 829: return "macaddr";
    case 869: return "inet";
    case 650: return "cidr";
    case 1042: return "bpchar";
    case 1043: return "varchar";
    case 1082: return "date";
    case 1083: return "time";
    case 1114: return "timestamp";
    case 1184: return "timestamptz";
    case 1186: return "interval";
    case 1266: return "timetz";
    case 1560: return "bit";
    case 1562: return "varbit";
    case 1700: return "numeric";
    default: return NULL;
  }
}

#define mondelefant_errcode_item(incode, outcode) \
  if (!strncmp(pgcode, (incode), strlen(incode))) return outcode; else

#define MONDELEFANT_ERRCODE_UNKNOWN "unknown"
#define MONDELEFANT_ERRCODE_CONNECTION "ConnectionException"
#define MONDELEFANT_ERRCODE_RESULTCOUNT_LOW "WrongResultSetCount.ResultSetMissing"
#define MONDELEFANT_ERRCODE_RESULTCOUNT_HIGH "WrongResultSetCount.TooManyResults"
#define MONDELEFANT_ERRCODE_QUERY1_NO_ROWS "NoData.OneRowExpected"
#define MONDELEFANT_ERRCODE_QUERY1_MULTIPLE_ROWS "CardinalityViolation.OneRowExpected"

static const char *mondelefant_translate_errcode(const char *pgcode) {
  if (!pgcode) abort();  // should not happen
  mondelefant_errcode_item("02", "NoData")
  mondelefant_errcode_item("03", "SqlStatementNotYetComplete")
  mondelefant_errcode_item("08", "ConnectionException")
  mondelefant_errcode_item("09", "TriggeredActionException")
  mondelefant_errcode_item("0A", "FeatureNotSupported")
  mondelefant_errcode_item("0B", "InvalidTransactionInitiation")
  mondelefant_errcode_item("0F", "LocatorException")
  mondelefant_errcode_item("0L", "InvalidGrantor")
  mondelefant_errcode_item("0P", "InvalidRoleSpecification")
  mondelefant_errcode_item("21", "CardinalityViolation")
  mondelefant_errcode_item("22", "DataException")
  mondelefant_errcode_item("23001", "IntegrityConstraintViolation.RestrictViolation")
  mondelefant_errcode_item("23502", "IntegrityConstraintViolation.NotNullViolation")
  mondelefant_errcode_item("23503", "IntegrityConstraintViolation.ForeignKeyViolation")
  mondelefant_errcode_item("23505", "IntegrityConstraintViolation.UniqueViolation")
  mondelefant_errcode_item("23514", "IntegrityConstraintViolation.CheckViolation")
  mondelefant_errcode_item("23",    "IntegrityConstraintViolation")
  mondelefant_errcode_item("24", "InvalidCursorState")
  mondelefant_errcode_item("25", "InvalidTransactionState")
  mondelefant_errcode_item("26", "InvalidSqlStatementName")
  mondelefant_errcode_item("27", "TriggeredDataChangeViolation")
  mondelefant_errcode_item("28", "InvalidAuthorizationSpecification")
  mondelefant_errcode_item("2B", "DependentPrivilegeDescriptorsStillExist")
  mondelefant_errcode_item("2D", "InvalidTransactionTermination")
  mondelefant_errcode_item("2F", "SqlRoutineException")
  mondelefant_errcode_item("34", "InvalidCursorName")
  mondelefant_errcode_item("38", "ExternalRoutineException")
  mondelefant_errcode_item("39", "ExternalRoutineInvocationException")
  mondelefant_errcode_item("3B", "SavepointException")
  mondelefant_errcode_item("3D", "InvalidCatalogName")
  mondelefant_errcode_item("3F", "InvalidSchemaName")
  mondelefant_errcode_item("40", "TransactionRollback")
  mondelefant_errcode_item("42", "SyntaxErrorOrAccessRuleViolation")
  mondelefant_errcode_item("44", "WithCheckOptionViolation")
  mondelefant_errcode_item("53", "InsufficientResources")
  mondelefant_errcode_item("54", "ProgramLimitExceeded")
  mondelefant_errcode_item("55", "ObjectNotInPrerequisiteState")
  mondelefant_errcode_item("57", "OperatorIntervention")
  mondelefant_errcode_item("58", "SystemError")
  mondelefant_errcode_item("F0", "ConfigurationFileError")
  mondelefant_errcode_item("P0", "PlpgsqlError")
  mondelefant_errcode_item("XX", "InternalError")
  return "unknown";
}

static int mondelefant_check_error_class(
  const char *errcode, const char *errclass
) {
  size_t i = 0;
  while (1) {
    if (errclass[i] == 0) {
      if (errcode[i] == 0 || errcode[i] == '.') return 1;
      else return 0;
    }
    if (errcode[i] != errclass[i]) return 0;
    i++;
  }
}

static void mondelefant_push_first_line(lua_State *L, const char *str) {
  char *str2;
  size_t i = 0;
  if (!str) abort();  // should not happen
  str2 = strdup(str);
  while (1) {
    char c = str2[i];
    if (c == '\n' || c == '\r' || c == 0) { str2[i] = 0; break; }
    i++;
  };
  lua_pushstring(L, str2);
  free(str2);
}

static int mondelefant_connect(lua_State *L) {
  luaL_Buffer buf;
  const char *conninfo;
  PGconn *pgconn;
  mondelefant_conn_t *conn;
  lua_settop(L, 1);
  lua_getfield(L, 1, "engine");  // 2
  if (!lua_toboolean(L, 2)) {
    return luaL_error(L, "No database engine selected.");
  }
  lua_pushliteral(L, "postgresql");  // 3
  if (!lua_rawequal(L, 2, 3)) {
    return luaL_error(L,
      "Only database engine 'postgresql' is supported."
    );
  }
  lua_settop(L, 1);
  lua_pushnil(L);  // slot for key at stack position 2
  lua_pushnil(L);  // slot for value at stack position 3
  luaL_buffinit(L, &buf);
  {
    int need_seperator = 0;
    while (lua_pushvalue(L, 2), lua_next(L, 1)) {
      lua_replace(L, 3);
      lua_replace(L, 2);
      // NOTE: numbers will be converted to strings automatically here,
      // but perhaps this will change in future versions of lua
      luaL_argcheck(L,
        lua_isstring(L, 2) && lua_isstring(L, 3), 1, "non-string contained"
      );
      lua_pushvalue(L, 2);
      lua_pushliteral(L, "engine");
      if (!lua_rawequal(L, -2, -1)) {
        const char *value;
        size_t value_len;
        size_t value_pos = 0;
        lua_pop(L, 1);
        if (need_seperator) luaL_addchar(&buf, ' ');
        luaL_addvalue(&buf);
        luaL_addchar(&buf, '=');
        luaL_addchar(&buf, '\'');
        value = lua_tolstring(L, 3, &value_len);
        do {
          char c;
          c = value[value_pos++];
          if (c == '\'') luaL_addchar(&buf, '\\');
          luaL_addchar(&buf, c);
        } while (value_pos < value_len);
        luaL_addchar(&buf, '\'');
        need_seperator = 1;
      } else {
        lua_pop(L, 1);
      }
    }
  }
  luaL_pushresult(&buf);
  lua_replace(L, 2);
  lua_settop(L, 2);
  conninfo = lua_tostring(L, 2);
  pgconn = PQconnectdb(conninfo);
  if (!pgconn) {
    return luaL_error(L,
      "Error in libpq while creating 'PGconn' structure."
    );
  }
  if (PQstatus(pgconn) != CONNECTION_OK) {
    const char *errmsg;
    lua_pushnil(L);
    errmsg = PQerrorMessage(pgconn);
    if (errmsg) {
      mondelefant_push_first_line(L, errmsg);
    } else {
      lua_pushliteral(L,
        "Error while connecting to database, but no error message given."
      );
    }
    lua_pushliteral(L, MONDELEFANT_ERRCODE_CONNECTION);
    PQfinish(pgconn);
    return 3;
  }
  lua_settop(L, 0);
  conn = lua_newuserdata(L, sizeof(*conn));  // 1
  conn->pgconn = pgconn;
  {
    const char *charset;
    charset = PQparameterStatus(pgconn, "server_encoding");
    if (charset && !strcmp(charset, "UTF8")) {
      conn->server_encoding = MONDELEFANT_SERVER_ENCODING_UTF8;
    } else {
      conn->server_encoding = MONDELEFANT_SERVER_ENCODING_ASCII;
    }
  }

  luaL_getmetatable(L, MONDELEFANT_CONN_MT_REGKEY);  // 2
  lua_setmetatable(L, 1);

  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CONN_DATA_REGKEY);  // 2
  lua_pushvalue(L, 1);  // 3
  lua_newtable(L);  // 4
  lua_settable(L, 2);
  lua_settop(L, 1);

  lua_pushliteral(L, "postgresql");
  lua_setfield(L, 1, "engine");
  return 1;
}

static mondelefant_conn_t *mondelefant_get_conn(lua_State *L, int index) {
  mondelefant_conn_t *conn;
  conn = luaL_checkudata(L, index, MONDELEFANT_CONN_MT_REGKEY);
  if (!conn->pgconn) {
    luaL_error(L, "PostgreSQL connection has been closed.");
    return NULL;
  }
  return conn;
}

static int mondelefant_conn_index(lua_State *L) {
  lua_settop(L, 2);
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CONN_DATA_REGKEY);  // 3
  lua_pushvalue(L, 1);  // 4
  lua_gettable(L, 3);  // 4
  lua_remove(L, 3);  // connection specific data-table at stack position 3
  lua_pushvalue(L, 2);  // 4
  lua_gettable(L, 3);  // 4
  if (!lua_isnil(L, 4)) return 1;
  lua_settop(L, 3);
  lua_getfield(L, 3, "prototype");  // 4
  if (lua_toboolean(L, 4)) {
    lua_pushvalue(L, 2);  // 5
    lua_gettable(L, 4);  // 5
    if (!lua_isnil(L, 5)) return 1;
  }
  lua_settop(L, 2);
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_MODULE_REGKEY);  // 3
  lua_getfield(L, 3, "postgresql_connection_prototype");  // 4
  if (lua_toboolean(L, 4)) {
    lua_pushvalue(L, 2);  // 5
    lua_gettable(L, 4);  // 5
    if (!lua_isnil(L, 5)) return 1;
  }
  lua_settop(L, 3);
  lua_getfield(L, 3, "connection_prototype");  // 4
  if (lua_toboolean(L, 4)) {
    lua_pushvalue(L, 2);  // 5
    lua_gettable(L, 4);  // 5
    if (!lua_isnil(L, 5)) return 1;
  }
  return 0;
}

static int mondelefant_conn_newindex(lua_State *L) {
  lua_settop(L, 3);
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CONN_DATA_REGKEY);  // 4
  lua_pushvalue(L, 1);  // 5
  lua_gettable(L, 4);  // 5
  lua_remove(L, 4);  // connection specific data-table  at stack position 4
  lua_pushvalue(L, 2);
  lua_pushvalue(L, 3);
  lua_settable(L, 4);
  return 0;
}

static int mondelefant_conn_free(lua_State *L) {
  mondelefant_conn_t *conn;
  conn = luaL_checkudata(L, 1, MONDELEFANT_CONN_MT_REGKEY);
  if (conn->pgconn) PQfinish(conn->pgconn);
  conn->pgconn = NULL;
  return 0;
}

static int mondelefant_conn_close(lua_State *L) {
  mondelefant_conn_t *conn;
  lua_settop(L, 1);
  conn = mondelefant_get_conn(L, 1);
  PQfinish(conn->pgconn);
  conn->pgconn = NULL;
  return 0;
}

static int mondelefant_conn_is_ok(lua_State *L) {
  mondelefant_conn_t *conn;
  lua_settop(L, 1);
  conn = mondelefant_get_conn(L, 1);
  lua_pushboolean(L, PQstatus(conn->pgconn) == CONNECTION_OK);
  return 1;
}

static int mondelefant_conn_get_transaction_status(lua_State *L) {
  mondelefant_conn_t *conn;
  lua_settop(L, 1);
  conn = mondelefant_get_conn(L, 1);
  switch (PQtransactionStatus(conn->pgconn)) {
  case PQTRANS_IDLE:
    lua_pushliteral(L, "idle");
    break;
  case PQTRANS_ACTIVE:
    lua_pushliteral(L, "active");
    break;
  case PQTRANS_INTRANS:
    lua_pushliteral(L, "intrans");
    break;
  case PQTRANS_INERROR:
    lua_pushliteral(L, "inerror");
    break;
  default:
    lua_pushliteral(L, "unknown");
  }
  return 1;
}

static int mondelefant_conn_create_list(lua_State *L) {
  lua_settop(L, 2);
  luaL_checkudata(L, 1, MONDELEFANT_CONN_MT_REGKEY);
  if (!lua_toboolean(L, 2)) {
    lua_newtable(L);
    lua_replace(L, 2);  // new result at stack position 2
  }
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_RESULT_MT_REGKEY);  // 3
  lua_setmetatable(L, 2);
  lua_pushvalue(L, 1);  // 3
  lua_setfield(L, 2, "_connection");
  lua_pushliteral(L, "list");  // 3
  lua_setfield(L, 2, "_type");
  return 1;
}

static int mondelefant_conn_create_object(lua_State *L) {
  lua_settop(L, 2);
  luaL_checkudata(L, 1, MONDELEFANT_CONN_MT_REGKEY);
  if (!lua_toboolean(L, 2)) {
    lua_newtable(L);
    lua_replace(L, 2);  // new result at stack position 2
  }
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_RESULT_MT_REGKEY);  // 3
  lua_setmetatable(L, 2);
  lua_pushvalue(L, 1);  // 3
  lua_setfield(L, 2, "_connection");
  lua_pushliteral(L, "object");  // 3
  lua_setfield(L, 2, "_type");  // "object" or "list"
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "_data");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "_dirty");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "_ref");  // nil=no info, false=nil, else table
  return 1;
}

static int mondelefant_conn_quote_string(lua_State *L) {
  mondelefant_conn_t *conn;
  const char *input;
  size_t input_len;
  char *output;
  size_t output_len;
  lua_settop(L, 2);
  conn = mondelefant_get_conn(L, 1);
  input = luaL_checklstring(L, 2, &input_len);
  if (input_len > (SIZE_MAX / sizeof(char) - 3) / 2) {
    return luaL_error(L, "String to be escaped is too long.");
  }
  output = malloc((2 * input_len + 3) * sizeof(char));
  if (!output) {
    return luaL_error(L, "Could not allocate memory for string quoting.");
  }
  output[0] = '\'';
  output_len = PQescapeStringConn(
    conn->pgconn, output + 1, input, input_len, NULL
  );
  output[output_len + 1] = '\'';
  output[output_len + 2] = 0;
  lua_pushlstring(L, output, output_len + 2);
  free(output);
  return 1;
}

static int mondelefant_conn_quote_binary(lua_State *L) {
  mondelefant_conn_t *conn;
  const char *input;
  size_t input_len;
  char *output;
  size_t output_len;
  luaL_Buffer buf;
  lua_settop(L, 2);
  conn = mondelefant_get_conn(L, 1);
  input = luaL_checklstring(L, 2, &input_len);
  output = (char *)PQescapeByteaConn(
    conn->pgconn, (const unsigned char *)input, input_len, &output_len
  );
  if (!output) {
    return luaL_error(L, "Could not allocate memory for binary quoting.");
  }
  luaL_buffinit(L, &buf);
  luaL_addchar(&buf, '\'');
  luaL_addlstring(&buf, output, output_len - 1);
  luaL_addchar(&buf, '\'');
  luaL_pushresult(&buf);
  PQfreemem(output);
  return 1;
}

static int mondelefant_conn_assemble_command(lua_State *L) {
  mondelefant_conn_t *conn;
  int paramidx = 2;
  const char *template;
  size_t template_pos = 0;
  luaL_Buffer buf;
  lua_settop(L, 2);
  conn = mondelefant_get_conn(L, 1);
  if (lua_isstring(L, 2)) {
    lua_tostring(L, 2);
    return 1;
  }
  // extra feature for objects with __tostring meta-method:
  if (luaL_callmeta(L, 2, "__tostring")) return 1;
  luaL_checktype(L, 2, LUA_TTABLE);
  lua_rawgeti(L, 2, 1);  // 3
  luaL_argcheck(L,
    lua_isstring(L, 3),
    2,
    "First entry of SQL command structure is not a string."
  );
  template = lua_tostring(L, 3);
  lua_pushliteral(L, "input_converter");  // 4
  lua_gettable(L, 1);  // input_converter at stack position 4
  lua_pushnil(L);  // free space at stack position 5
  lua_pushnil(L);  // free space at stack position 6
  luaL_buffinit(L, &buf);
  while (1) {
    char c;
    c = template[template_pos++];
    if (!c) break;
    if (c == '?' || c == '$') {
      if (template[template_pos] == c) {
        template_pos++;
        luaL_addchar(&buf, c);
      } else {
        luaL_Buffer keybuf;
        int subcmd;
        subcmd = (c == '$');
        luaL_buffinit(L, &keybuf);
        while (1) {
          c = template[template_pos];
          if (
            (c < 'A' || c > 'Z') &&
            (c < 'a' || c > 'z') &&
            (c < '0' || c > '9') &&
            (c != '_')
          ) break;
          luaL_addchar(&keybuf, c);
          template_pos++;
        }
        luaL_pushresult(&keybuf);
        if (lua_objlen(L, -1)) {
          lua_pushvalue(L, -1);           // save key on stack
          lua_gettable(L, 2);             // fetch value
        } else {
          lua_pop(L, 1);
          lua_pushnil(L);                 // put nil on key position
          lua_rawgeti(L, 2, paramidx++);  // fetch value
        }
        // stack: ..., <buffer>, key, pre-value
        if (subcmd) {
          size_t i;
          size_t count;
          lua_replace(L, 5);  // sub-structure at stack position 5
          lua_pop(L, 1);      // drop stored key
          // stack: ..., <buffer>
          luaL_argcheck(L,
            !lua_isnil(L, 5),
            2,
            "SQL sub-structure not found."
          );
          luaL_argcheck(L,
            lua_type(L, 5) == LUA_TTABLE,
            2,
            "SQL sub-structure must be a table."
          );
          // stack: ..., <buffer>
          lua_getfield(L, 5, "sep");
          lua_replace(L, 6);  // seperator at stack position 6
          if (lua_isnil(L, 6)) {
            lua_pushstring(L, ", ");
            lua_replace(L, 6);
          } else {
            luaL_argcheck(L,
              lua_isstring(L, 6),
              2,
              "Seperator of SQL sub-structure has to be a string."
            );
          }
          count = lua_objlen(L, 5);
          for (i = 0; i < count; i++) {
            if (i) {
              lua_pushvalue(L, 6);
              luaL_addvalue(&buf);
            }
            lua_pushcfunction(L, mondelefant_conn_assemble_command);
            lua_pushvalue(L, 1);
            lua_rawgeti(L, 5, i+1);
            lua_call(L, 2, 1);
            luaL_addvalue(&buf);
          }
        } else {
          if (lua_toboolean(L, 4)) {
            // call input_converter with connection handle, value and info
            lua_pushvalue(L, 4);
            lua_pushvalue(L, 1);
            lua_pushvalue(L, -3);
            lua_newtable(L);
            lua_pushvalue(L, -6);
            lua_setfield(L, -2, "field_name");
            lua_call(L, 3, 1);
            // stack: ..., <buffer>, key, pre-value, final-value
            lua_remove(L, -2);
            lua_remove(L, -2);
            // stack: ..., <buffer>, final-value
            if (!lua_isstring(L, -1)) {
              return luaL_error(L, "input_converter returned non-string.");
            }
          } else {
            lua_remove(L, -2);
            // stack: ..., <buffer>, pre-value
            if (lua_isnil(L, -1)) {
              lua_pushliteral(L, "NULL");
            } else if (lua_type(L, -1) == LUA_TBOOLEAN) {
              lua_pushstring(L, lua_toboolean(L, -1) ? "TRUE" : "FALSE");
            } else if (lua_isstring(L, -1)) {
              // NOTE: In this version of lua a number will be converted
              lua_tostring(L, -1);
              lua_pushcfunction(L, mondelefant_conn_quote_string);
              lua_pushvalue(L, 1);
              lua_pushvalue(L, -3);
              lua_call(L, 2, 1);
            } else {
              return luaL_error(L,
                "Unable to convert SQL value due to unknown type "
                "or missing input_converter."
              );
            }
            // stack: ..., <buffer>, pre-value, final-value
            lua_remove(L, -2);
            // stack: ..., <buffer>, final-value
          }
          luaL_addvalue(&buf);
        }
      }
    } else {
      luaL_addchar(&buf, c);
    }
  }
  luaL_pushresult(&buf);
  return 1;
}

#define MONDELEFANT_MAX_COMMAND_COUNT 64
#define MONDELEFANT_MAX_COLUMN_COUNT 1024
#define MONDELEFANT_QUERY_MODE_LIST 1
#define MONDELEFANT_QUERY_MODE_OBJECT 2
#define MONDELEFANT_QUERY_MODE_OPT_OBJECT 3

static int mondelefant_conn_try_query(lua_State *L) {
  mondelefant_conn_t *conn;
  int command_count;
  int command_idx;
  int modes[MONDELEFANT_MAX_COMMAND_COUNT];
  luaL_Buffer buf;
  int sent_success;
  PGresult *res;
  int rows, cols, row, col;
  conn = mondelefant_get_conn(L, 1);
  command_count = lua_gettop(L) / 2;
  lua_pushnil(L);  // needed, if last mode was omitted
  if (command_count > MONDELEFANT_MAX_COMMAND_COUNT) {
    return luaL_error(L, "Exceeded maximum command count in one query.");
  }
  luaL_buffinit(L, &buf);
  for (command_idx = 0; command_idx < command_count; command_idx++) {
    int mode;
    int mode_idx;  // stack index of mode string
    if (command_idx) luaL_addchar(&buf, ' ');
    lua_pushcfunction(L, mondelefant_conn_assemble_command);
    lua_pushvalue(L, 1);
    lua_pushvalue(L, 2 + 2 * command_idx);
    lua_call(L, 2, 1);
    luaL_addvalue(&buf);
    luaL_addchar(&buf, ';');
    mode_idx = 3 + 2 * command_idx;
    if (lua_isnil(L, mode_idx)) {
      mode = MONDELEFANT_QUERY_MODE_LIST;
    } else {
      const char *modestr;
      modestr = luaL_checkstring(L, mode_idx);
      if (!strcmp(modestr, "list")) {
        mode = MONDELEFANT_QUERY_MODE_LIST;
      } else if (!strcmp(modestr, "object")) {
        mode = MONDELEFANT_QUERY_MODE_OBJECT;
      } else if (!strcmp(modestr, "opt_object")) {
        mode = MONDELEFANT_QUERY_MODE_OPT_OBJECT;
      } else {
        return luaL_error(L, "Unknown query mode specified.");
      }
    }
    modes[command_idx] = mode;
  }
  luaL_pushresult(&buf);  // stack position unknown
  lua_replace(L, 2);  // SQL command string to stack position 2
  lua_settop(L, 2);
  lua_getfield(L, 1, "sql_tracer");  // tracer at stack position 3
  if (lua_toboolean(L, 3)) {
    lua_pushvalue(L, 1);  // 4
    lua_pushvalue(L, 2);  // 5
    lua_call(L, 2, 1);  // trace callback at stack position 3
  }
  sent_success = PQsendQuery(conn->pgconn, lua_tostring(L, 2));
  lua_newtable(L);  // results in table at stack position 4
  for (command_idx = 0; ; command_idx++) {
    int mode;
    char binary[MONDELEFANT_MAX_COLUMN_COUNT];
    ExecStatusType pgstatus;
    mode = modes[command_idx];
    if (sent_success) {
      res = PQgetResult(conn->pgconn);
      if (command_idx >= command_count && !res) break;
      if (res) {
        pgstatus = PQresultStatus(res);
        rows = PQntuples(res);
        cols = PQnfields(res);
      }
    }
    if (
      !sent_success || command_idx >= command_count || !res ||
      (pgstatus != PGRES_TUPLES_OK && pgstatus != PGRES_COMMAND_OK) ||
      (rows < 1 && mode == MONDELEFANT_QUERY_MODE_OBJECT) ||
      (rows > 1 && mode != MONDELEFANT_QUERY_MODE_LIST)
    ) {
      const char *command;
      command = lua_tostring(L, 2);
      lua_newtable(L);  // 5
      lua_getfield(L,
        LUA_REGISTRYINDEX,
        MONDELEFANT_ERROROBJECT_MT_REGKEY
      );
      lua_setmetatable(L, 5);
      lua_pushvalue(L, 1);
      lua_setfield(L, 5, "connection");
      lua_pushinteger(L, command_idx + 1);
      lua_setfield(L, 5, "command_number");
      lua_pushvalue(L, 2);
      lua_setfield(L, 5, "sql_command");
      if (!res) {
        lua_pushliteral(L, MONDELEFANT_ERRCODE_RESULTCOUNT_LOW);
        lua_setfield(L, 5, "code");
        lua_pushliteral(L, "Received too few database result sets.");
        lua_setfield(L, 5, "message");
      } else if (command_idx >= command_count) {
        lua_pushliteral(L, MONDELEFANT_ERRCODE_RESULTCOUNT_HIGH);
        lua_setfield(L, 5, "code");
        lua_pushliteral(L, "Received too many database result sets.");
        lua_setfield(L, 5, "message");
      } else if (
        pgstatus != PGRES_TUPLES_OK && pgstatus != PGRES_COMMAND_OK
      ) {
        const char *sqlstate;
        const char *errmsg;
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_SEVERITY));
        lua_setfield(L, 5, "pg_severity");
        sqlstate = PQresultErrorField(res, PG_DIAG_SQLSTATE);
        if (sqlstate) {
          lua_pushstring(L, sqlstate);
          lua_setfield(L, 5, "pg_sqlstate");
          lua_pushstring(L, mondelefant_translate_errcode(sqlstate));
          lua_setfield(L, 5, "code");
        } else {
          lua_pushliteral(L, MONDELEFANT_ERRCODE_UNKNOWN);
          lua_setfield(L, 5, "code");
        }
        errmsg = PQresultErrorField(res, PG_DIAG_MESSAGE_PRIMARY);
        if (errmsg) {
          mondelefant_push_first_line(L, errmsg);
          lua_setfield(L, 5, "message");
          lua_pushstring(L, errmsg);
          lua_setfield(L, 5, "pg_message_primary");
        } else {
          lua_pushliteral(L,
            "Error while fetching result, but no error message given."
          );
          lua_setfield(L, 5, "message");
        }
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_MESSAGE_DETAIL));
        lua_setfield(L, 5, "pg_message_detail");
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_MESSAGE_HINT));
        lua_setfield(L, 5, "pg_message_hint");
        // NOTE: "position" and "pg_internal_position" are recalculated to
        // byte offsets, as Lua 5.1 is not Unicode aware.
        {
          char *tmp;
          tmp = PQresultErrorField(res, PG_DIAG_STATEMENT_POSITION);
          if (tmp) {
            int pos;
            pos = atoi(tmp) - 1;
            if (conn->server_encoding == MONDELEFANT_SERVER_ENCODING_UTF8) {
              pos = utf8_position_to_byte(command, pos);
            }
            lua_pushinteger(L, pos + 1);
            lua_setfield(L, 5, "position");
          }
        }
        {
          const char *internal_query;
          internal_query = PQresultErrorField(res, PG_DIAG_INTERNAL_QUERY);
          lua_pushstring(L, internal_query);
          lua_setfield(L, 5, "pg_internal_query");
          char *tmp;
          tmp = PQresultErrorField(res, PG_DIAG_INTERNAL_POSITION);
          if (tmp) {
            int pos;
            pos = atoi(tmp) - 1;
            if (conn->server_encoding == MONDELEFANT_SERVER_ENCODING_UTF8) {
              pos = utf8_position_to_byte(internal_query, pos);
            }
            lua_pushinteger(L, pos + 1);
            lua_setfield(L, 5, "pg_internal_position");
          }
        }
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_CONTEXT));
        lua_setfield(L, 5, "pg_context");
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_SOURCE_FILE));
        lua_setfield(L, 5, "pg_source_file");
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_SOURCE_LINE));
        lua_setfield(L, 5, "pg_source_line");
        lua_pushstring(L, PQresultErrorField(res, PG_DIAG_SOURCE_FUNCTION));
        lua_setfield(L, 5, "pg_source_function");
      } else if (rows < 1 && mode == MONDELEFANT_QUERY_MODE_OBJECT) {
        lua_pushliteral(L, MONDELEFANT_ERRCODE_QUERY1_NO_ROWS);
        lua_setfield(L, 5, "code");
        lua_pushliteral(L, "Expected one row, but got empty set.");
        lua_setfield(L, 5, "message");
      } else if (rows > 1 && mode != MONDELEFANT_QUERY_MODE_LIST) {
        lua_pushliteral(L, MONDELEFANT_ERRCODE_QUERY1_MULTIPLE_ROWS);
        lua_setfield(L, 5, "code");
        lua_pushliteral(L, "Got more than one result row.");
        lua_setfield(L, 5, "message");
      } else {
        // should not happen
        abort();
      }
      if (res) {
        PQclear(res);
        while ((res = PQgetResult(conn->pgconn))) PQclear(res);
      }
      if (lua_toboolean(L, 3)) {
        lua_pushvalue(L, 3);
        lua_pushvalue(L, 5);
        lua_call(L, 1, 0);
      }
      return 1;
    }
    rows = PQntuples(res);
    cols = PQnfields(res);
    if (modes[command_idx] == MONDELEFANT_QUERY_MODE_LIST) {
      lua_pushcfunction(L, mondelefant_conn_create_list);  // 5
      lua_pushvalue(L, 1);  // 6
      lua_call(L, 1, 1);  // 5
    } else {
      lua_pushcfunction(L, mondelefant_conn_create_object);  // 5
      lua_pushvalue(L, 1);  // 6
      lua_call(L, 1, 1);  // 5
    }
    lua_newtable(L);  // column_info at atack position 6
    for (col = 0; col < cols; col++) {
      lua_newtable(L);  // 7
      lua_pushstring(L, PQfname(res, col));
      lua_setfield(L, 7, "field_name");
      {
        Oid tmp;
        tmp = PQftable(res, col);
        if (tmp == InvalidOid) lua_pushnil(L);
        else lua_pushinteger(L, tmp);
        lua_setfield(L, 7, "table_oid");
      }
      {
        int tmp;
        tmp = PQftablecol(res, col);
        if (tmp == 0) lua_pushnil(L);
        else lua_pushinteger(L, tmp);
        lua_setfield(L, 7, "table_column_number");
      }
      {
        Oid tmp;
        tmp = PQftype(res, col);
        binary[col] = (tmp == MONDELEFANT_POSTGRESQL_BINARY_OID);
        lua_pushinteger(L, tmp);
        lua_setfield(L, 7, "type_oid");
        lua_pushstring(L, mondelefant_oid_to_typestr(tmp));
        lua_setfield(L, 7, "type");
      }
      {
        int tmp;
        tmp = PQfmod(res, col);
        if (tmp == -1) lua_pushnil(L);
        else lua_pushinteger(L, tmp);
        lua_setfield(L, 7, "type_modifier");
      }
      lua_rawseti(L, 6, col+1);
    }
    lua_setfield(L, 5, "_column_info");  // stack at position 5 with result
    {
      char *tmp;
      tmp = PQcmdTuples(res);
      if (tmp[0]) {
        lua_pushinteger(L, atoi(tmp));
        lua_setfield(L, 5, "_rows_affected");
      }
    }
    {
      Oid tmp;
      tmp = PQoidValue(res);
      if (tmp != InvalidOid) {
        lua_pushinteger(L, tmp);
        lua_setfield(L, 5, "_oid");
      }
    }
    if (modes[command_idx] == MONDELEFANT_QUERY_MODE_LIST) {
      for (row = 0; row < rows; row++) {
        lua_pushcfunction(L, mondelefant_conn_create_object);  // 6
        lua_pushvalue(L, 1);  // 7
        lua_call(L, 1, 1);  // 6
        for (col = 0; col < cols; col++) {
          if (PQgetisnull(res, row, col)) {
            lua_pushnil(L);
          } else if (binary[col]) {
            size_t binlen;
            char *binval;
            binval = (char *)PQunescapeBytea(
              (unsigned char *)PQgetvalue(res, row, col), &binlen
            );
            if (!binval) {
              return luaL_error(L,
                "Could not allocate memory for binary unescaping."
              );
            }
            lua_pushlstring(L, binval, binlen);
            PQfreemem(binval);
          } else {
            lua_pushstring(L, PQgetvalue(res, row, col));
          }
          lua_rawseti(L, 6, col+1);
        }
        lua_rawseti(L, 5, row+1);
      }
    } else if (rows == 1) {
      for (col = 0; col < cols; col++) {
        if (PQgetisnull(res, 0, col)) {
          lua_pushnil(L);
        } else if (binary[col]) {
          size_t binlen;
          char *binval;
          binval = (char *)PQunescapeBytea(
            (unsigned char *)PQgetvalue(res, 0, col), &binlen
          );
          if (!binval) {
            return luaL_error(L,
              "Could not allocate memory for binary unescaping."
            );
          }
          lua_pushlstring(L, binval, binlen);
          PQfreemem(binval);
        } else {
          lua_pushstring(L, PQgetvalue(res, 0, col));
        }
        lua_rawseti(L, 5, col+1);
      }
    } else {
      // no row in optrow mode
      lua_pop(L, 1);
      lua_pushnil(L);
    }
    lua_rawseti(L, 4, command_idx+1);
    if (lua_gettop(L) != 4) abort();  // should not happen
    PQclear(res);
  }
  // trace callback at stack position 3
  // result at stack position 4 (top of stack)
  if (lua_toboolean(L, 3)) {
    lua_pushvalue(L, 3);
    lua_call(L, 0, 0);
  }
  lua_replace(L, 3);  // result at stack position 3
  lua_getfield(L, 1, "output_converter");  // output converter at stack position 4
  for (command_idx = 0; command_idx < command_count; command_idx++) {
    int mode;
    mode = modes[command_idx];
    lua_rawgeti(L, 3, command_idx+1);  // raw result at stack position 5
    if (lua_toboolean(L, 5)) {
      lua_getfield(L, 5, "_column_info");  // column_info list at position 6
      cols = lua_objlen(L, 6);
      if (mode == MONDELEFANT_QUERY_MODE_LIST) {
        rows = lua_objlen(L, 5);
        for (row = 0; row < rows; row++) {
          lua_rawgeti(L, 5, row+1);  // row at stack position 7
          lua_getfield(L, 7, "_data");  // _data table at stack position 8
          for (col = 0; col < cols; col++) {
            lua_rawgeti(L, 6, col+1);  // this column info at position 9
            lua_getfield(L, 9, "field_name");  // 10
            if (lua_toboolean(L, 4)) {
              lua_pushvalue(L, 4);  // output-converter
              lua_pushvalue(L, 1);  // connection
              lua_rawgeti(L, 7, col+1);  // raw-value
              lua_pushvalue(L, 9);  // this column info
              lua_call(L, 3, 1);  // converted value at position 11
            } else {
              lua_rawgeti(L, 7, col+1);  // raw-value at position 11
            }
            lua_pushvalue(L, 11);  // 12
            lua_rawseti(L, 7, col+1);
            lua_rawset(L, 8);
            lua_settop(L, 8);
          }
          lua_settop(L, 6);
        }
      } else {
        lua_getfield(L, 5, "_data");  // _data table at stack position 7
        for (col = 0; col < cols; col++) {
          lua_rawgeti(L, 6, col+1);  // this column info at position 8
          lua_getfield(L, 8, "field_name");  // 9
          if (lua_toboolean(L, 4)) {
            lua_pushvalue(L, 4);  // output-converter
            lua_pushvalue(L, 1);  // connection
            lua_rawgeti(L, 5, col+1);  // raw-value
            lua_pushvalue(L, 8);  // this column info
            lua_call(L, 3, 1);  // converted value at position 10
          } else {
            lua_rawgeti(L, 5, col+1);  // raw-value at position 10
          }
          lua_pushvalue(L, 10);  // 11
          lua_rawseti(L, 5, col+1);
          lua_rawset(L, 7);
          lua_settop(L, 7);
        }
      }
    }
    lua_settop(L, 4);
  }
  lua_settop(L, 3);
  lua_pushnil(L);
  for (command_idx = 0; command_idx < command_count; command_idx++) {
    lua_rawgeti(L, 3, command_idx+1);
  }
  return command_count+1;
}

static int mondelefant_errorobject_escalate(lua_State *L) {
  lua_settop(L, 1);
  lua_getfield(L, 1, "connection");  // 2
  lua_getfield(L, 2, "error_objects");  // 3
  if (lua_toboolean(L, 3)) {
    lua_settop(L, 1);
    return lua_error(L);
  } else {
    lua_getfield(L, 1, "message");  // 4
    if (lua_isnil(L, 4)) {
      return luaL_error(L, "No error message given for escalation.");
    }
    return lua_error(L);
  }
}

static int mondelefant_errorobject_is_kind_of(lua_State *L) {
  lua_settop(L, 2);
  lua_getfield(L, 1, "code");  // 3
  if (lua_isstring(L, 3)) {
    lua_pushboolean(L,
      mondelefant_check_error_class(
        lua_tostring(L, 3), luaL_checkstring(L, 2)
      )
    );
  } else {
    // only happens for errors where code is not set
    lua_pushboolean(L, 0);
  }
  return 1;
}

static int mondelefant_conn_query(lua_State *L) {
  int argc;
  argc = lua_gettop(L);
  lua_pushvalue(L, 1);
  lua_insert(L, 1);
  lua_pushcfunction(L, mondelefant_conn_try_query);
  lua_insert(L, 2);
  lua_call(L, argc, LUA_MULTRET);  // results (with error) starting at index 2
  if (lua_toboolean(L, 2)) {
    lua_pushcfunction(L, mondelefant_errorobject_escalate);
    lua_pushvalue(L, 2);
    lua_call(L, 1, 0);  // will raise an error
    return 0;  // should not be executed
  } else {
    return lua_gettop(L) - 2;
  }
}

static int mondelefant_set_class(lua_State *L) {
  lua_settop(L, 2);
  lua_getmetatable(L, 1);  // 3
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_RESULT_MT_REGKEY);  // 4
  luaL_argcheck(L, lua_equal(L, 3, 4), 1, "not a database result");
  lua_settop(L, 2);
  lua_getmetatable(L, 2);  // 3
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CLASS_MT_REGKEY);  // 4
  luaL_argcheck(L, lua_equal(L, 3, 4), 2, "not a database class");
  lua_settop(L, 2);
  lua_pushvalue(L, 2);  // 3
  lua_setfield(L, 1, "_class");
  lua_getfield(L, 1, "_type");  // 3
  lua_pushliteral(L, "list");  // 4
  if (lua_rawequal(L, 3, 4)) {
    int i;
    for (i=0; i < lua_objlen(L, 1); i++) {
      lua_settop(L, 2);
      lua_rawgeti(L, 1, i+1);  // 3
      lua_pushvalue(L, 2);  // 4
      lua_setfield(L, 3, "_class");
    }
  }
  lua_settop(L, 1);
  return 1;
}

static int mondelefant_new_class(lua_State *L) {
  lua_settop(L, 1);
  if (!lua_toboolean(L, 1)) {
    lua_settop(L, 0);
    lua_newtable(L);  // 1
  }
  lua_getfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CLASS_MT_REGKEY);  // 2
  lua_setmetatable(L, 1);
  lua_pushliteral(L, "prototype");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "prototype");  // 3
    lua_getfield(L,
      LUA_REGISTRYINDEX,
      MONDELEFANT_CLASS_PROTO_REGKEY
    );  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "object");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "object");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "object_get");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "object_get");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "object_set");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "object_set");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "list");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "list");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "references");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "references");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  lua_pushliteral(L, "foreign_keys");  // 2
  lua_rawget(L, 1);  // 2
  if (!lua_toboolean(L, 2)) {
    lua_pushliteral(L, "foreign_keys");  // 3
    lua_newtable(L);  // 4
    lua_rawset(L, 1);
  }
  lua_settop(L, 1);
  return 1;
}

static int mondelefant_class_get_reference(lua_State *L) {
  lua_settop(L, 2);
  while (lua_toboolean(L, 1)) {
    lua_getfield(L, 1, "references");  // 3
    lua_pushvalue(L, 2);  // 4
    lua_gettable(L, 3);  // 4
    if (!lua_isnil(L, 4)) return 1;
    lua_settop(L, 2);
    lua_pushliteral(L, "prototype");  // 3
    lua_rawget(L, 1);  // 3
    lua_replace(L, 1);
  }
  return 0;
}

static int mondelefant_class_iterate_over_references(lua_State *L) {
  return luaL_error(L, "Reference iterator not implemented yet.");  // TODO
}

static int mondelefant_class_get_foreign_key_reference_name(lua_State *L) {
  lua_settop(L, 2);
  while (lua_toboolean(L, 1)) {
    lua_getfield(L, 1, "foreign_keys");  // 3
    lua_pushvalue(L, 2);  // 4
    lua_gettable(L, 3);  // 4
    if (!lua_isnil(L, 4)) return 1;
    lua_settop(L, 2);
    lua_pushliteral(L, "prototype");  // 3
    lua_rawget(L, 1);  // 3
    lua_replace(L, 1);
  }
  return 0;
}

static int mondelefant_result_index(lua_State *L) {
  const char *result_type;
  lua_settop(L, 2);
  if (lua_type(L, 2) != LUA_TSTRING || lua_tostring(L, 2)[0] == '_') {
    lua_rawget(L, 1);
    return 1;
  }
  lua_getfield(L, 1, "_class");  // 3
  if (!lua_toboolean(L, 3)) {
    lua_settop(L, 2);
    lua_getfield(L,
      LUA_REGISTRYINDEX,
      MONDELEFANT_CLASS_PROTO_REGKEY
    );  // 3
  }
  lua_getfield(L, 1, "_type");  // 4
  result_type = lua_tostring(L, 4);
  if (result_type && !strcmp(result_type, "object")) {
    lua_settop(L, 3);
    // try inherited attributes, methods or getter functions:
    while (lua_toboolean(L, 3)) {
      lua_getfield(L, 3, "object");  // 4
      lua_pushvalue(L, 2);  // 5
      lua_gettable(L, 4);  // 5
      if (!lua_isnil(L, 5)) return 1;
      lua_settop(L, 3);
      lua_getfield(L, 3, "object_get");  // 4
      lua_pushvalue(L, 2);  // 5
      lua_gettable(L, 4);  // 5
      if (lua_toboolean(L, 5)) {
        lua_pushvalue(L, 1);  // 6
        lua_call(L, 1, 1);  // 5
        return 1;
      }
      lua_settop(L, 3);
      lua_pushliteral(L, "prototype");  // 4
      lua_rawget(L, 3);  // 4
      lua_replace(L, 3);
    }
    lua_settop(L, 2);
    // try primary keys of referenced objects:
    lua_pushcfunction(L,
      mondelefant_class_get_foreign_key_reference_name
    );  // 3
    lua_getfield(L, 1, "_class");  // 4
    lua_pushvalue(L, 2);  // 5
    lua_call(L, 2, 1);  // 3
    if (!lua_isnil(L, 3)) {
      // reference name at stack position 3
      lua_pushcfunction(L, mondelefant_class_get_reference);  // 4
      lua_getfield(L, 1, "_class");  // 5
      lua_pushvalue(L, 3);  // 6
      lua_call(L, 2, 1);  // reference info at stack position 4
      lua_getfield(L, 1, "_ref");  // 5
      lua_getfield(L, 4, "ref");  // 6
      lua_gettable(L, 5);  // 6
      if (!lua_isnil(L, 6)) {
        if (lua_toboolean(L, 6)) {
          lua_getfield(L, 4, "that_key");  // 7
          if (lua_isnil(L, 7)) {
            return luaL_error(L, "Missing 'that_key' entry in model reference.");
          }
          lua_gettable(L, 6);  // 7
        } else {
          lua_pushnil(L);
        }
        return 1;
      }
    }
    lua_settop(L, 2);
    // try normal data field info:
    lua_getfield(L, 1, "_data");  // 3
    lua_pushvalue(L, 2);  // 4
    lua_gettable(L, 3);  // 4
    if (!lua_isnil(L, 4)) return 1;
    lua_settop(L, 2);
    // try cached referenced object (or cached NULL reference):
    lua_getfield(L, 1, "_ref");  // 3
    lua_pushvalue(L, 2);  // 4
    lua_gettable(L, 3);  // 4
    if (lua_isboolean(L, 4) && !lua_toboolean(L, 4)) {
      lua_pushnil(L);
      return 1;
    } else if (!lua_isnil(L, 4)) {
      return 1;
    }
    lua_settop(L, 2);
    // try to load a referenced object:
    lua_pushcfunction(L, mondelefant_class_get_reference);  // 3
    lua_getfield(L, 1, "_class");  // 4
    lua_pushvalue(L, 2);  // 5
    lua_call(L, 2, 1);  // 3
    if (!lua_isnil(L, 3)) {
      lua_settop(L, 2);
      lua_getfield(L, 1, "load");  // 3
      lua_pushvalue(L, 1);  // 4 (self)
      lua_pushvalue(L, 2);  // 5
      lua_call(L, 2, 0);
      lua_settop(L, 2);
      lua_getfield(L, 1, "_ref");  // 3
      lua_pushvalue(L, 2);  // 4
      lua_gettable(L, 3);  // 4
      if (lua_isboolean(L, 4) && !lua_toboolean(L, 4)) lua_pushnil(L);  // TODO: use special object instead of false
      return 1;
    }
    return 0;
  } else if (result_type && !strcmp(result_type, "list")) {
    lua_settop(L, 3);
    // try inherited list attributes or methods:
    while (lua_toboolean(L, 3)) {
      lua_getfield(L, 3, "list");  // 4
      lua_pushvalue(L, 2);  // 5
      lua_gettable(L, 4);  // 5
      if (!lua_isnil(L, 5)) return 1;
      lua_settop(L, 3);
      lua_pushliteral(L, "prototype");  // 4
      lua_rawget(L, 3);  // 4
      lua_replace(L, 3);
    }
  }
  return 0;
}

static int mondelefant_result_newindex(lua_State *L) {
  const char *result_type;
  lua_settop(L, 3);
  if (lua_type(L, 2) != LUA_TSTRING || lua_tostring(L, 2)[0] == '_') {
    lua_rawset(L, 1);
    return 1;
  }
  lua_getfield(L, 1, "_class");  // 4
  if (!lua_toboolean(L, 4)) {
    lua_settop(L, 3);
    lua_getfield(L,
      LUA_REGISTRYINDEX,
      MONDELEFANT_CLASS_PROTO_REGKEY
    );  // 4
  }
  lua_getfield(L, 1, "_type");  // 5
  result_type = lua_tostring(L, 5);
  if (result_type && !strcmp(result_type, "object")) {
    lua_settop(L, 4);
    // try object setter functions:
    while (lua_toboolean(L, 4)) {
      lua_getfield(L, 4, "object_set");  // 5
      lua_pushvalue(L, 2);  // 6
      lua_gettable(L, 5);  // 6
      if (lua_toboolean(L, 6)) {
        lua_pushvalue(L, 1);  // 7
        lua_pushvalue(L, 3);  // 8
        lua_call(L, 2, 0);
        return 0;
      }
      lua_settop(L, 4);
      lua_pushliteral(L, "prototype");  // 5
      lua_rawget(L, 4);  // 5
      lua_replace(L, 4);
    }
    lua_settop(L, 3);
    // check, if a object reference is changed:
    lua_pushcfunction(L, mondelefant_class_get_reference);  // 4
    lua_getfield(L, 1, "_class");  // 5
    lua_pushvalue(L, 2);  // 6
    lua_call(L, 2, 1);  // 4
    if (!lua_isnil(L, 4)) {
      // store object in _ref table (use false for nil):  // TODO: use special object instead of false
      lua_getfield(L, 1, "_ref");  // 5
      lua_pushvalue(L, 2);  // 6
      if (lua_isnil(L, 3)) lua_pushboolean(L, 0);  // 7
      else lua_pushvalue(L, 3);  // 7
      lua_settable(L, 5);
      lua_settop(L, 4);
      // delete referencing key from _data table:
      lua_getfield(L, 4, "this_key");  // 5
      if (lua_isnil(L, 5)) {
        return luaL_error(L, "Missing 'this_key' entry in model reference.");
      }
      lua_getfield(L, 1, "_data");  // 6
      lua_pushvalue(L, 5);  // 7
      lua_pushnil(L);  // 8
      lua_settable(L, 6);
      lua_settop(L, 5);
      lua_getfield(L, 1, "_dirty");  // 6
      lua_pushvalue(L, 5);  // 7
      lua_pushboolean(L, 1);  // 8
      lua_settable(L, 6);
      return 0;
    }
    lua_settop(L, 3);
    // store value in data field info:
    lua_getfield(L, 1, "_data");  // 4
    lua_pushvalue(L, 2);  // 5
    lua_pushvalue(L, 3);  // 6
    lua_settable(L, 4);
    lua_settop(L, 3);
    // mark field as dirty (needs to be UPDATEd on save):
    lua_getfield(L, 1, "_dirty");  // 4
    lua_pushvalue(L, 2);  // 5
    lua_pushboolean(L, 1);  // 6
    lua_settable(L, 4);
    lua_settop(L, 3);
    // reset reference cache, if neccessary:
    lua_pushcfunction(L,
      mondelefant_class_get_foreign_key_reference_name
    );  // 4
    lua_getfield(L, 1, "_class");  // 5
    lua_pushvalue(L, 2);  // 6
    lua_call(L, 2, 1);  // 4
    if (!lua_isnil(L, 4)) {
      lua_getfield(L, 1, "_ref");  // 5
      lua_pushvalue(L, 4);  // 6
      lua_pushnil(L);  // 7
      lua_settable(L, 5);
    }
    return 0;
  } else {
    lua_settop(L, 3);
    lua_rawset(L, 1);
    return 0;
  }
  return 0;
}

static int mondelefant_class_index(lua_State *L) {
  lua_settop(L, 2);
  lua_pushliteral(L, "prototype");  // 3
  lua_rawget(L, 1);  // 3
  lua_pushvalue(L, 2);  // 4
  lua_gettable(L, 3);  // 4
  return 1;
}

static const struct luaL_Reg mondelefant_module_functions[] = {
  {"connect", mondelefant_connect},
  {"set_class", mondelefant_set_class},
  {"new_class", mondelefant_new_class},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_conn_mt_functions[] = {
  {"__gc", mondelefant_conn_free},
  {"__index", mondelefant_conn_index},
  {"__newindex", mondelefant_conn_newindex},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_conn_methods[] = {
  {"close", mondelefant_conn_close},
  {"is_ok", mondelefant_conn_is_ok},
  {"get_transaction_status", mondelefant_conn_get_transaction_status},
  {"create_list", mondelefant_conn_create_list},
  {"create_object", mondelefant_conn_create_object},
  {"quote_string", mondelefant_conn_quote_string},
  {"quote_binary", mondelefant_conn_quote_binary},
  {"assemble_command", mondelefant_conn_assemble_command},
  {"try_query", mondelefant_conn_try_query},
  {"query", mondelefant_conn_query},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_errorobject_mt_functions[] = {
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_errorobject_methods[] = {
  {"escalate", mondelefant_errorobject_escalate},
  {"is_kind_of", mondelefant_errorobject_is_kind_of},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_result_mt_functions[] = {
  {"__index", mondelefant_result_index},
  {"__newindex", mondelefant_result_newindex},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_class_mt_functions[] = {
  {"__index", mondelefant_class_index},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_class_methods[] = {
  {"get_reference", mondelefant_class_get_reference},
  {"iterate_over_references", mondelefant_class_iterate_over_references},
  {"get_foreign_key_reference_name",
    mondelefant_class_get_foreign_key_reference_name},
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_object_methods[] = {
  {NULL, NULL}
};

static const struct luaL_Reg mondelefant_list_methods[] = {
  {NULL, NULL}
};

int luaopen_mondelefant_native(lua_State *L) {
  lua_settop(L, 0);
  lua_newtable(L);  // module at stack position 1
  luaL_register(L, NULL, mondelefant_module_functions);

  lua_pushvalue(L, 1);  // 2
  lua_setfield(L, LUA_REGISTRYINDEX, MONDELEFANT_MODULE_REGKEY);

  lua_newtable(L);  // 2
  // NOTE: only PostgreSQL is supported yet:
  luaL_register(L, NULL, mondelefant_conn_methods);
  lua_setfield(L, 1, "postgresql_connection_prototype");
  lua_newtable(L);  // 2
  lua_setfield(L, 1, "connection_prototype");

  luaL_newmetatable(L, MONDELEFANT_CONN_MT_REGKEY);  // 2
  luaL_register(L, NULL, mondelefant_conn_mt_functions);
  lua_settop(L, 1);
  luaL_newmetatable(L, MONDELEFANT_RESULT_MT_REGKEY);  // 2
  luaL_register(L, NULL, mondelefant_result_mt_functions);
  lua_setfield(L, 1, "result_metatable");
  luaL_newmetatable(L, MONDELEFANT_CLASS_MT_REGKEY);  // 2
  luaL_register(L, NULL, mondelefant_class_mt_functions);
  lua_setfield(L, 1, "class_metatable");

  lua_newtable(L);  // 2
  lua_newtable(L);  // 3
  luaL_register(L, NULL, mondelefant_object_methods);
  lua_setfield(L, 2, "object");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "object_get");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "object_set");
  lua_newtable(L);  // 3
  luaL_register(L, NULL, mondelefant_list_methods);
  lua_setfield(L, 2, "list");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "references");
  lua_newtable(L);  // 3
  lua_setfield(L, 2, "foreign_keys");
  lua_pushvalue(L, 2);  // 3
  lua_setfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CLASS_PROTO_REGKEY);
  lua_setfield(L, 1, "class_prototype");

  lua_newtable(L);  // 2
  lua_pushliteral(L, "k");  // 3
  lua_setfield(L, 2, "__mode");
  lua_newtable(L);  // 3
  lua_pushvalue(L, 2);  // 4
  lua_setmetatable(L, 3);
  lua_setfield(L, LUA_REGISTRYINDEX, MONDELEFANT_CONN_DATA_REGKEY);
  lua_settop(L, 1);

  luaL_newmetatable(L, MONDELEFANT_ERROROBJECT_MT_REGKEY);  // 2
  luaL_register(L, NULL, mondelefant_errorobject_mt_functions);
  lua_newtable(L);  // 3
  luaL_register(L, NULL, mondelefant_errorobject_methods);
  lua_setfield(L, 2, "__index");
  lua_setfield(L, 1, "errorobject_metatable");

  return 1;
}
